"""
Overview:
    Design pattern module. Even in python, design patterns are not emphasized like java, \
    but some simple packages are often useful, such as singleton patterns.
"""
from .decorator import *
from .final import *
from .observer import *
from .singleton import *
